#include "processor.h"
#include "command.cc"
#include "evaluate.cc"

const bool Y86::init = init_table();                   // инициализация таблицы операций

class outstr : public ostringstream {                  // поток вывода дампов
public:
   outstr( void ) {
      setf( ios::hex, ios::basefield );
      fill( '0' );   
   }
};

//------------------------- диагностика --------------------------------
string Y86::current_line( void ) { 
   try {
      return at( IP );
   }
   catch( out_of_range const& e ) {
      outstr oret;
      oret << "несуществующий указатель команды: " << "0x" 
           << setw( ADRLEN * 2 ) << IP;
	  return oret.str();
   }  
}

string Y86::get_state( void ) {
   outstr oret;
   oret << "Executed " << dec << count << " steps at IP = " 
        << "0x" << setw( ADRLEN * 2 ) << hex  << IP << ". Status "
        << ( HALT == state ? "HALT" : RUN == state ? "RUNNING" : "LOADING" )
        << ", ZF=" << ( ZF ? "1" : "0" )
        << " SF=" << ( SF ? "1" : "0" )
        << " OF=" << ( OF ? "1" : "0" );
   return oret.str();
}

static const string regname[ Y86::rnum ] = { 
   "%rax", "%rcx", "%rdx", "%rbx",
   "%rsp", "%rbp", "%rsi", "%rdi",
   "%r8 ", "%r9 ", "%r10", "%r11",
   "%r12", "%r13", "%r14" 
};

string Y86::get_regs( void ) {
   outstr oret;
   oret << "Registers:" << endl;
   int lin = ( rnum + 1 ) / 2;
   for( int i = 0; i < lin; i++ ) { 
      oret << regname[ i ] << "=0x" << setw( w2b * 2 ) << R[ i ];
      if( i + lin < rnum ) 
         oret << "        " << regname[ i + lin ] << "=0x"  
              << setw( w2b * 2 ) << R[ i + lin ];
      oret << endl;
   }
   return oret.str();
}

string Y86::dump( void ) {
   outstr oret;
   oret << "Memory:" << endl;
   for( uint32_t addr = 0; addr < ram.size(); addr++ ) {
      if( 0 == addr % 16 )
         oret << "0x" << setw( ADRLEN * 2 ) << addr << ":  ";
      oret << setw( 2 ) << (int)ram[ addr ] << " ";
      if( 15 == addr % 16 ) oret << endl;
   }
   if( ram.size() %16 != 0 ) oret << endl;
   return oret.str();
}

string Y86::compare( Y86& p ) {
   outstr oret;
   oret << "Changes to registers:" << endl;
   for( int i = 0; i < rnum; i++ )
      if( R[ i ] != p.R[ i ] ) 
         oret << regname[ i ] << ":   0x" << setw( w2b * 2 ) << p.R[ i ] 
              << "      0x" << setw( w2b * 2 ) << R[ i ] << endl;
   oret << endl << "Changes to memory:" << endl;
   for( uint32_t addr = 0; addr < ram.size(); addr += 4 ) {
	  uint32_t v1 = value_at( addr ),
	           v2 = p.value_at( addr );  
      if( v1 != v2 ) 
         oret << "0x" << setw( ADRLEN * 2 ) << addr 
              << ": 0x" << setw( w2b * 2 ) << v2 
              << "      0x" << setw( w2b * 2 ) << v1 << endl; 

   }
   return oret.str();
}

//--------------------- обмен словами с памятью ------------------------

void Y86::save64( uint64_t val, addr_t addr ) {        // сохранить 64-значение в память  
   uint8_t vb[ w2b ];
   for( int i = 0; i < w2b; val >>= 8, i++ )
      vb[ i ] = val & 0xff;
   if( !little_endian )                                // изменение порядка байт      
      reverse( vb, vb + w2b ); 
   for( int i = 0; i < w2b; i++ )    
      ram[ addr + i ] = vb[ i ];
}

uint64_t Y86::value_at( addr_t addr ) {                // считать значение из памяти
   uint64_t vr = 0;
   uint8_t  vb[ w2b ];
   for( int i = 0; i < w2b; i++ )
      vb[ i ] = ram[ addr + i ];
   if( little_endian )                                 // изменение порядка байт      
      reverse( vb, vb + w2b );
   for( int i = 0; i < w2b; i++ ) {
      vr += vb[ i ];
      if( i < w2b - 1 ) vr <<= 8;
   }
   return vr;
}

//----------------------------------------------------------------------

bool Y86::step( void ) {                               // выполнение одной команды
   if( HALT == state ) {
      cout << "программа завершена" << endl;
	  return true;      
   }
   for( int i = 0; i < ncomnd; i++ ) 
      if( comnds[ i ].cod == ram[ IP ] ) {
		 state = RUN;
		 command* pc = (command*)comnds[ i ].func;
		 (*pc)( this, comnds + i );
         count++;
		 return true;
      }
   cerr << "недопустимый код команды: 0x" << hex << ram[ IP ] << endl; 
   return false;
}

//----------------------------------------------------------------------

void Y86::load( addr_t addr, const vector<int>& ops ) {
   if( debug_level > 1 ) {
          cout << "addr=0x" << hex << setw( ADRLEN * 2 )
               << setfill( '0' ) << addr << "\t::\t";
      for( auto x : ops )
         cout << hex << setw( 2 ) << setfill( '0' ) << (int)x << " ";
      cout << endl;
   }
   for( auto x : ops ) 
      try {
         ram.at( addr ) = x;
         addr++;
      }
      catch( out_of_range& e ) {                       // выход за предел текущей памяти
         addr_t lim = w2b * ( ( addr + ops.size() ) / w2b +
                              ( ( addr + ops.size() ) % w2b != 0 ? 1 : 0 ) ); 
         ram.resize( lim, 0 );                         // расширить до границы слова
         if( debug_level > 1 )
            cout << "расширение памяти до : " << hex << setw( 4 )
                 << setfill( '0' ) << ram.capacity() << endl;
         ram.at( addr++ ) = x;
      }
}

istream& operator >>( istream& inp, Y86& pr ) {
   if( debug_level > 1 )
      cout << "начальный размер памяти : " << pr.ram.capacity() << endl;
   unsigned maxaddr = 0;
   while( true ) {
      getline( inp, pr.line );
      if( inp.eof() ) break;
      string::size_type found = pr.line.find( '|' );
      if( string::npos == found ) return inp;          // не ассемблированная строка?
      string code = pr.line.substr( 0, found );        // строка программы
      found = code.find_last_not_of( ' ' );            // удалить хвостовые пробелы
      if( found != string::npos )
         code = code.substr( 0, found + 1 );
      else
         continue;                                     // пустая срока 
      found = code.find( ":" );
      if( string::npos == found ) continue;            // это может быть комментарий 
      string saddr = code.substr( 0, found ), sdata;
      if( code.length() - found > 1 )
         sdata = code.substr( found + 2 );
      addr_t addr; 
      try {
         vector<int> ops;
         addr = stol( saddr, NULL, 16 );
         for( unsigned i = 0; i < sdata.length(); i += 2 ) {
            string s = sdata.substr( i, 2 );
            int byte = stoi( s, NULL, 16 );
            ops.push_back( byte );
         }
         if( !ops.empty() ) {                          // не директива
            pr.load( addr, ops );
            pr.insert( pair< addr_t, string >( addr, pr.line ) );
         }
      }
      catch( exception const& e ) {                    // ошибка записи кода
         return inp;
      }
      maxaddr = addr > maxaddr ? addr : maxaddr;  
   }
   if( pr.ram.size() < maxaddr ) {
      addr_t lim = w2b * ( maxaddr / w2b + ( maxaddr % w2b != 0 ? 1 : 0 ) ); 
	  pr.ram.resize( lim, 0 );
   } 
   pr.line.clear();
   pr.OK = true;                                       // обнулить процессор
   pr.IP = 0;
   pr.count = 0;
   for( auto &r : pr.R ) r = 0;
   pr.state = pr.LOAD;
   pr.ZF = pr.SF = pr.OF = false;
   if( debug_level > 1 ) {	   
      for( auto x : pr.ram )
         cout <<  hex << setw( 2 ) << setfill( '0' ) << (int)x << " ";
      cout << endl;
   }
   if( debug_level > 0 )
      cout << "размер памяти программы : 0x" <<  hex << setw( 8 ) 
           << setfill( '0' ) << pr.ram.size() << endl;
   return inp;
}
